#ifndef SYMBOL_H
#define SYMBOL_H

#include "expr.h"	/* needed for 'struct Expr' */
#include "ida.h"
#include "typedef.h"

/*{{{  constant definitions*/
/*{{{  'sp->attrs' flags  ('AT_...') -- classes & attrs*/
/*------------------------------------------------------*/
/* Flags in 'sp->attrs'   ('AT_...')			*/
/*							*/
/* N.B. This assumes that 'int32' is 'long'		*/
/*------------------------------------------------------*/

/*{{{  simple flags*/
/*--------------*/
/* entity types */
/*--------------*/
/*-- variables --*/

#define  AT_ALLOC		0x1L	/* allocatable array (must be array) */
#define  AT_POINTER		0x2L	/* pointer (scalar or array) */
#define  AT_NORMAL_VAR		0x4L	/* variable that isn't one of above */
#define  AT_STMT_VAR		0x8L	/* statement entity dummy variable */

/*-- non variables --*/

#define  AT_PARAM		0x10L	/* named constant */

#define  AT_FUNC		0x20L
#define  AT_SUBR		0x40L
#define  AT_MODULE		0x80L	/* a MODULE: the hostunit or USEd */
#define  AT_PROGRAM		0x100L	/* the PROGRAM unit */
#define  AT_BLOCK_DATA		0x200L	/* a BLOCK DATA unit */

#define  AT_DERIVED_TYPE	0x400L	/* derived-type name */
#define  AT_NAMELIST_GROUP	0x800L
#define  AT_CONSTRUCT		0x1000L

/*-- HPF entities --*/

#define  AT_PROCESSORS		0x2000L	/* HPF processors (scalar or array)*/
#define  AT_TEMPLATE		0x4000L	/* HPF template (scalar or array) */

/*----------------------------------------------------------------
N.B. We exclude AT_COMMON_BLOCK from the composite flag AT_ENTITY,
since it can coexist with most of the other entity flags, as a
common block can have the same name as any local object except a
named const, intrinsic proc or result var!
----------------------------------------------------------------*/
#define  AT_COMMON_BLOCK	0x8000L
#define  AT_COMMON_SAVE		0x10000L   /* for saved COMMON block     */
					   /* (AT_COMMON_BLOCK also set) */
/*-------------------------*/
/* certain storage classes */
/*-------------------------*/
#define  AT_DUMMY		0x20000L   /* dummy argument */
#define  AT_RESULT_VAR		0x40000L   /* function result variable */
#define  AT_IN_COMMON		0x80000L   /* common block variable */
#define  AT_EQUIV		0x100000L  /* appearing in an EQUIVALENCE set */

/*------------*/
/* attributes */
/*------------*/
#define  AT_EXTERNAL		0x200000L  /* declared EXTERNAL */
#define  AT_INTRINSIC		0x400000L  /* declared INTRINSIC */
#define  AT_IN			0x800000L  /* dummy defined as 'IN'  */
#define  AT_OUT			0x1000000L /* dummy defined as 'OUT' */
#define  AT_OPTIONAL		0x2000000L /* optional dummy argument */
#define  AT_PRIVATE		0x4000000L
#define  AT_PUBLIC		0x8000000L
#define  AT_SAVE		0x10000000L
#define  AT_TARGET		0x20000000L

/*-------------------*/
/* other information */
/*-------------------*/
#define  AT_IN_NAMELIST		0x40000000L /* appearing in a NAMELIST group */
#define  AT_INFO		0x80000000L /* 'sp->info' contains other info */


/*}}}*/
/*{{{  aliases for simple flags*/

#define  AT_INIT	AT_INFO    /* statically initialised variable */
/*}}}*/

/*--------------------------------------*/
/* Composite flags for 'sp->attrs'	*/
/*--------------------------------------*/
/*{{{  composite flags (i.e. groups of flags)*/
/*--------------------------------------------------------------*/
/* AT_ENTITY excludes AT_COMMON_BLOCK, as a common block can	*/
/* have the same name as any local object except a named	*/
/* constant or intrinsic procedure!				*/
/*--------------------------------------------------------------*/
#define  AT_ENTITY	(AT_ALLOC | AT_POINTER | AT_NORMAL_VAR | AT_STMT_VAR |\
			 AT_PARAM | AT_FUNC | AT_SUBR | AT_MODULE |\
			 AT_PROGRAM | AT_BLOCK_DATA | AT_DERIVED_TYPE |\
			 AT_NAMELIST_GROUP | AT_CONSTRUCT |\
			 AT_PROCESSORS | AT_TEMPLATE)
			 
/*--------------------------------------------------------------*/
/* AT_ATTRS -- any attribute that can appear in a typedecl,	*/
/*             except those that declare 'entity types'		*/
/*             (i.e. ALLOC, POINTER, PARAM).			*/
/*--------------------------------------------------------------*/
#define  AT_ATTRS	(AT_EXTERNAL | AT_IN | AT_OUT | AT_INTRINSIC |\
			 AT_OPTIONAL | AT_PRIVATE | AT_PUBLIC | AT_SAVE |\
			 AT_TARGET)

/*-- composite attrs --*/

#define  AT_ACCESS		(AT_PRIVATE | AT_PUBLIC)
#define  AT_INOUT		(AT_IN | AT_OUT)
#define  AT_INTENT		AT_INOUT

#define  AT_COMMON_NAME		(AT_COMMON_BLOCK | AT_COMMON_SAVE)
					/* appears as a common block name */

#define  AT_INTRIN_FUNC		(AT_INTRINSIC | AT_RESULT_VAR)
#define  AT_INTRIN_SUBR		(AT_INTRINSIC | AT_SUBR)
#define  AT_DUMMY_IN		(AT_DUMMY | AT_IN)
#define  AT_DUMMY_OUT		(AT_DUMMY | AT_OUT)
#define  AT_DUMMY_INOUT		(AT_DUMMY | AT_INOUT)
#define  AT_OPT_DUMMY_IN	(AT_DUMMY | AT_IN | AT_OPTIONAL)
#define  AT_OPT_DUMMY_OUT	(AT_DUMMY | AT_OUT | AT_OPTIONAL)

/*--------------------------------------------------------------*/
/* AT_PROCEDURE -- a procedure, however declared (viz., host	*/
/* unit, explicit interface, INTRINSIC or EXTERNAL stmts or	*/
/* attrs).							*/
/*								*/
/* N.B. Including 'AT_EXTERNAL' here isn't quite correct, as it	*/
/* *could* be a BLOCK DATA rather than a procedure.  Is this ok?*/
/*--------------------------------------------------------------*/
#define  AT_PROCEDURE		(AT_EXTERNAL | AT_INTRINSIC |\
				 AT_FUNC | AT_SUBR)

/*{{{  AT_DATA_OBJECT,  AT_NONDATA_ENTITY*/
/*--------------------------------------------------------------*/
/* AT_DATA_OBJECT    -- a variable or named constant.		*/
/* AT_NONDATA_ENTITY -- *not* a data object.			*/
/*								*/
/* N.B. AT_ENTITY == (AT_DATA_OBJECT | AT_NONDATA_ENTITY)	*/
/*	These groups exclude AT_COMMON_BLOCK, as this can have	*/
/* the same name as a data object or another nondata entity!	*/
/*--------------------------------------------------------------*/
#define  AT_DATA_OBJECT		(AT_ALLOC | AT_POINTER | AT_NORMAL_VAR |\
				 AT_STMT_VAR | AT_PARAM)

#define  AT_NONDATA_ENTITY	(AT_FUNC | AT_SUBR | AT_MODULE | AT_PROGRAM |\
				 AT_BLOCK_DATA | AT_DERIVED_TYPE |\
				 AT_NAMELIST_GROUP | AT_CONSTRUCT |\
				 AT_PROCESSORS | AT_TEMPLATE)
/*}}}*/

/*{{{  AT_TYPED_ENTITY,  AT_UNTYPED_ENTITY*/
/*--------------------------------------------------------------*/
/* AT_TYPED_ENTITY   -- any entity with a datatype.		*/
/* AT_UNTYPED_ENTITY -- any entity *without* a datatype.	*/
/*								*/
/* N.B. AT_ENTITY == (AT_TYPED_ENTITY | AT_UNTYPED_ENTITY)	*/
/*	These groups exclude AT_COMMON_BLOCK, as this can have	*/
/* the same name as a typed entity or another untyped entity!	*/
/*--------------------------------------------------------------*/
#define  AT_TYPED_ENTITY	(AT_DATA_OBJECT | AT_FUNC)

#define  AT_UNTYPED_ENTITY	(AT_SUBR | AT_MODULE | AT_PROGRAM |\
				 AT_BLOCK_DATA | AT_DERIVED_TYPE |\
				 AT_NAMELIST_GROUP | AT_CONSTRUCT |\
				 AT_PROCESSORS | AT_TEMPLATE)
/*}}}*/

/*{{{  AT_VAR,  AT_NONVAR*/
/*--------------------------------------------------------------*/
/* AT_VAR          -- a variable.				*/
/* AT_NONVAR       -- *not* a variable.				*/
/*								*/
/* N.B.								*/
/* ----								*/
/* --  (AT_VAR | AT_NONVAR) == (AT_ENTITY | AT_EXTERNAL|	*/
/*                              AT_INTRINSIC)			*/
/* --  After the decls section, if there are no AT_NONVAR	*/
/* flags set then AT_NORMAL_VAR can be safely set -- there's no	*/
/* danger of conflicting with any other 'attrs' flags.		*/
/* --  AT_NONVAR excludes AT_COMMON_BLOCK, as this can have	*/
/* the same name as a local or dummy variable!			*/
/*--------------------------------------------------------------*/
#define  AT_VAR			(AT_ALLOC | AT_POINTER | AT_NORMAL_VAR |\
				 AT_STMT_VAR)

#define  AT_NONVAR		(AT_PARAM | AT_FUNC | AT_SUBR | AT_MODULE |\
				 AT_PROGRAM | AT_BLOCK_DATA |\
				 AT_DERIVED_TYPE | AT_NAMELIST_GROUP |\
				 AT_CONSTRUCT | AT_PROCESSORS |\
				 AT_TEMPLATE | AT_EXTERNAL | AT_INTRINSIC)
/*}}}*/
/*}}}*/



/*}}}*/
/*{{{  'sp->type'  flags  ('TY_...') -- datatypes  (also used in 'ep->types')*/
/*------------------------------------------------------*/
/* Flags in 'sp->type'   ('TY_...')			*/
/* N.B. This assumes that 'int32' is 'long'		*/
/*							*/
/* Simple flags						*/
/*------------------------------------------------------*/

/*{{{  datatype flags   *** SOME USED IN 'ep->type' AS WELL AS 'sp->type' ****/
/*--------------------------------------------------------------*/
/* Datatype flags.  *** SOME OF THESE ARE USED IN 'ep->type' &	*/
/* 'ep->mapp->attrs' AS WELL AS HERE.  IN THAT CASE THEY'RE	*/
/* RENAMED 'EX_...' & 'MA_...'. ***				*/
/*	The lowest order bits (masked off by TY_KIND_MASK) are	*/
/* reserved for storing the KIND parameter, if any.		*/
/*								*/
/* N.B:  The TY_TYPE mask includes the KIND value field and the	*/
/* type flags, but excludes the TY_IMPLICIT flag.		*/
/*	The TY_BASIC_TYPE mask only includes the type flags,	*/
/* not the KIND value field.					*/
/*--------------------------------------------------------------*/

/****************************************************************/
/* N.B. THE FOLLOWING FLAGS ARE ALSO USED IN 'ep->type'	&	*/
/* 'ep->mapp->attrs' (RENAMED 'EX_...' & 'MA_...' RESPECTIVELY).*/
/* IF THEIR VALUES ARE CHANGED, THE VALUES OF ALL OTHER 'EX_...'*/
/* & 'MA_...' FLAGS MUST ALSO BE CHANGED IN "expr.h" TO AVOID	*/
/* CLASHES OF FLAG VALUES THEREIN.				*/
/****************************************************************/
/*-- mask for KIND parameter field --*/

#define  TY_KIND_MASK		0x3FL	/* ...can store KIND values < 64 */

/*-- data types --*/

#define  TY_INTEGER		0x40L
#define	 TY_REAL		0x80L
#define  TY_DP			0x100L
#define  TY_COMPLEX		0x200L
#define  TY_LOGICAL		0x400L
#define  TY_CHARACTER		0x800L
#define  TY_DERIVED		0x1000L
#define  TY_NONE		0x2000L

/*--------------------------------------------------------------*/
/* These and all subsequent flags are only used in 'sp->types'	*/
/*--------------------------------------------------------------*/
#define  TY_IMPLICIT		0x4000L

#define  TY_SEQUENCE		0x8000L  /* a sequence derived type */
#define  TY_PRIVATE_COMPNTS	0x10000L /* typedef with private compnts */
/*}}}*/

/*{{{  intrinsic proc types & classes  (used only in 'sp->types')*/
/*--------------------------------------------------------------*/
/* Type specs for intrinsic function results.			*/
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~			*/
/* (i) The result type is determined by the actual arg types:	*/
/*--------------------------------------------------------------*/
#define  TY_TYPE_ARG_1		    0x20000L   /* same type/kind as arg 1 */
#define  TY_TYPE_ARG_1_NONCOMPLEX   0x40000L   /* same type/kind as arg 1,
				                  except COMPLEX -> REAL */
#define  TY_TYPE_ARG_2		    0x80000L   /* same type/kind as arg 2.
						  (Only for TRANSFER fn.) */
#define  TY_TYPE_ARG_1_TIMES_ARG_2  0x100000L  /* same type/kind as
						  (arg1 * arg2) or
				                  (arg1 .AND. arg2).
						  (Only MATMUL/DOT_PRODUCT)*/

/*--------------------------------------------------------------*/
/* (ii) The result has a fixed *type* but its *kind* parameter	*/
/* is given by a 'KIND' argument if present, or is default	*/
/* otherwise.  This applies to just 8 type conversion functions:*/
/* -- 6 for numeric type conversion, perhaps also changing	*/
/*	value;							*/
/* -- 1 character, for converting int -> char;			*/
/* -- 1 logical, for logical kind conversion;			*/
/* viz:								*/
/*	AINT, ANINT, CMPLX, INT, NINT, REAL, CHAR, LOGICAL	*/
/*								*/
/* This flag always used in conjunction with an 'TY_type' flag	*/
/* to give a composite flag (e.g. TY_INT_KIND, TY_REAL_KIND...)	*/
/*--------------------------------------------------------------*/
#define  TY_KIND_ARG		0x200000L


/*--------------------------------------------------------------*/
/* Intrinsic procedure class					*/
/* ~~~~~~~~~~~~~~~~~~~~~~~~~					*/
/* i.e. elemental, inquiry or transformational;  generic or	*/
/* specific;  and in latter case, can it be used as an actual	*/
/* argument?							*/
/*--------------------------------------------------------------*/
#define  TY_ELEMENTAL		0x400000L	/* elemental fn/subr */
#define  TY_INQUIRY		0x800000L	/* inquiry fn */
#define  TY_SYS_INQUIRY		0x1000000L	/* (HPF) system inquiry fn */
#define  TY_TRANSF		0x2000000L	/* transformational fn */

#define  TY_GENERIC		0x4000000L	/* generic name */
#define  TY_SPECIFIC		0x8000000L	/* specific name */
#define  TY_ACTUAL		0x10000000L	/* can be used as actual arg */
/*}}}*/

/*{{{  proc attributes (RECURSIVE, PURE, EXTRINSIC)  (used only in 'sp->types')*/
/*----------------------*/
/* Procedure attributes	*/
/*----------------------*/
#define  TY_RECURSIVE		0x20000000L
#define  TY_PURE		0x40000000L	/* only for HPF */
#define  TY_EXTRINSIC		0x80000000L	/* only for HPF */
/*}}}*/

/*--------------------------------------*/
/* Composite flags for 'sp->type'	*/
/*--------------------------------------*/

/*{{{  composite flags*/

#define  TY_BASIC_TYPE	(TY_INTEGER | TY_REAL | TY_DP | TY_COMPLEX |\
			 TY_LOGICAL | TY_CHARACTER | TY_DERIVED | TY_NONE)

#define  TY_TYPE	(TY_KIND_MASK | TY_BASIC_TYPE)

/*-- composite types --*/

#define  TY_NUMERIC	(TY_INTEGER | TY_REAL | TY_DP | TY_COMPLEX)

/*--------------------------------------------------------------*/
/* Intrinsic function result type determined by the type(s) of	*/
/* actual argument(s).						*/
/*--------------------------------------------------------------*/
#define  TY_TYPE_ARG	(TY_TYPE_ARG_1 | TY_TYPE_ARG_1_NONCOMPLEX |\
			 TY_TYPE_ARG_2 | TY_TYPE_ARG_1_TIMES_ARG_2)

/*--------------------------------------------------------------*/
/* These composite types are used for intrinsic function dummy	*/
/* arguments -- they give the range of permissible datatypes.	*/
/*--------------------------------------------------------------*/
#define  TY_REAL_DP		(TY_REAL | TY_DP)
#define  TY_REAL_DP_COMPLEX	(TY_REAL | TY_DP | TY_COMPLEX)
#define  TY_INT_REAL_DP		(TY_INTEGER | TY_REAL | TY_DP)
#define  TY_NUM_LOGICAL		(TY_NUMERIC | TY_LOGICAL)

/*--------------------------------------------------------------*/
/* Intrinsic function result of definite *type*, whose kind	*/
/* parameter is given by a 'KIND' argument if it's present,	*/
/* or is default otherwise.					*/
/*--------------------------------------------------------------*/
#define  TY_INT_KIND		(TY_INTEGER | TY_KIND_ARG)
#define  TY_REAL_KIND		(TY_REAL | TY_KIND_ARG)
#define  TY_COMPLEX_KIND	(TY_COMPLEX | TY_KIND_ARG)
#define  TY_LOGICAL_KIND	(TY_LOGICAL | TY_KIND_ARG)
#define  TY_CHAR_KIND		(TY_CHARACTER | TY_KIND_ARG)

/*--------------------------------------------------------------*/
/* This is used for intrinsic function names that that are both	*/
/* specific *and* generic!  					*/
/*								*/
/* N.B: Occasionally, these can't be used as actual args	*/
/* (e.g. CHAR(I)).  In that case I wonder why F90 bothers with	*/
/* specific form!						*/
/*--------------------------------------------------------------*/
#define  TY_GEN_SPEC		(TY_GENERIC | TY_SPECIFIC)

/*}}}*/
/*}}}*/
/*{{{  'arp->attrs' flags  (often '..._SHAPE' or '..._SIZE')*/
/*-----------------------*/
/* Flags in 'arp->attrs' */
/*-----------------------*/
#define  RANK_MASK		0xF	/* has to take vals <= (MAX_RANK+1) */

#define  CONST_SHAPE		0x10	/* all upper bounds const */
#define  VAR_SHAPE		0x20	/* all upper bounds given, some var */
#define  ASSUMED_SIZE		0x40	/* last bound is '*' (-> CONST / VAR)*/
#define  ASSUMED_SHAPE		0x80	/* all upper bounds omitted, but some
					               lower bounds explicit */
#define  DEFERRED_SHAPE		0x100	/* all bounds omitted (alloc or ptr) */

#define  ASSOC_PTR_SHAPE	0x200	/* shape assumed from pointer assoc */

/*--------------------------------------------------------------*/
/* The following flags denote illegal shape specs.  They may	*/
/* only be set during the processing of a 'shape_spec', after	*/
/* which they're transformed to a standard shape.		*/
/*--------------------------------------------------------------*/
#define  MIXED_SHAPE		0x400	/* some dims assumed, some explicit */
#define  MIXED_ASSUMED_SIZE	0x800	/* '*' with assumed/mixed shape */
#define  INVALID_POSN_ASSUMED_SIZE					\
				0x1000	/* '*' not in final dim */


/*-- The next flag is synonymous with EXPL_SHAPE but only uses 1 bit!
It isn't set in 'arp->attrs', but is used in some functions that refer
to the array type (along with ASSUMED_SHAPE & ASSOC_PTR_SHAPE). --*/

#define  CONTIGUOUS		0x2000


/*-- The next flags are only for the dummy result vars of
transformational intrinsic functions --*/

#define  AR_DIM_ARG		0x4000	/* for reduction funcs & MAX/MINLOC*/
#define  AR_TRANSFER_FUNC	0x8000	/* only for 'TRANSER' intrinsic */


/*-----------------*/
/* Composite flags */
/*-----------------*/
#define  EXPL_SHAPE		(CONST_SHAPE | VAR_SHAPE)

#define  SHAPE_MASK		(CONST_SHAPE | VAR_SHAPE | ASSUMED_SIZE |\
				 ASSUMED_SHAPE | DEFERRED_SHAPE |\
				 ASSOC_PTR_SHAPE |\
				 MIXED_SHAPE | MIXED_ASSUMED_SIZE |\
				 INVALID_POSN_ASSUMED_SIZE)

#define  INVALID_SHAPE		(MIXED_SHAPE | MIXED_ASSUMED_SIZE |\
				 INVALID_POSN_ASSUMED_SIZE)
/*}}}*/

/*{{{  'alignp->attrs' flags ('AL_...')*/
/*----------------------------------------------*/
/* Flags for 'alignp->attrs'			*/
/*						*/
/* N.B. This assumes that 'int32' is 'long'	*/
/*----------------------------------------------*/
/*{{{  AL_SRC_ / _TGT_ / _EMBED_ / _REPLIC_RANK_MASK / _SHIFT*/
/*--------------------------------------------------------------*/
/* The following fields must accommodate values <= MAX_RANK,	*/
/* i.e. 7 with current Fortran rules, so they're 3 bits wide.	*/
/*--------------------------------------------------------------*/
#define  AL_SRC_RANK_MASK	0x7L		/* field for 'src_rank' */
#define  AL_TGT_RANK_MASK	0x38L		/*   "    "  'tgt_rank' */
#define  AL_EMBED_RANK_MASK	0x1C0L		/*   "    "  'embed_rank'*/
#define  AL_REPLIC_RANK_MASK	0xE00L		/*   "    "  'replic_rank'*/

/*----------------------------------------------------------------
'xxx_SHIFT' is the number of bits that a value must be shifted
(left) to reach the field in which it's stored in 'attrs'.
E.g., 'tgt_rank' is stored and read as follows:
   to store:	attrs   += tgt_rank << AL_TGT_RANK_SHIFT
   to read:	tgt_rank = (attrs & AL_TGT_RANK_MASK) >> AL_TGT_RANK_SHIFT
----------------------------------------------------------------*/
#define  AL_TGT_RANK_SHIFT	3		/* shift for 'tgt_rank' */
#define  AL_EMBED_RANK_SHIFT	6		/*   "    "  'embed_rank'*/
#define  AL_REPLIC_RANK_SHIFT	9		/*   "    "  'replic_rank'*/
/*}}}*/

/*{{{  AL_NEED_SRC_ / _TGT_LB / _MASK*/
/*----------------------------------------------------------------
	Flags '(AL_NEED_SRC_LB << dim)' and '(AL_NEED_TGT_LB << dim)',
0 <= dim < MAX_RANK,  are set if 'alignp->l [dim]' doesn't have
the final result because the lower bound of the source or target
respectively were unknown when the alignment was specified.
	'AL_NEED_SRC_LB_MASK' is the mask for the set of flags:
	     {(AL_NEED_SRC_LB << dim),  0 <= dim < MAX_RANK},
and likewise for 'AL_NEED_TGT_LB_MASK'.
----------------------------------------------------------------*/
#define  AL_NEED_SRC_LB		0x1000L
#define  AL_NEED_SRC_LB_MASK	0x7F000L

#define  AL_NEED_TGT_LB		0x80000L
#define  AL_NEED_TGT_LB_MASK	0x3F80000L
/*}}}*/

/*{{{  AL_EXPL_ / _DEDUCED_SRC_ / TGT_LIST*/
/*----------------------------------------------------------------
	Flags 'AL_EXPL_SRC_/TGT_LIST' mean that the source and/or
target list(s) was (were) explicitly specified in ALIGN.

	Only one of the flags 'AL_DEDUCED_SRC_/TGT_LIST' may be set.
It means that the corresponding list was omitted from ALIGN, so it
must be a list of (:,:,...), and its length was *deduced* from the
explicit list or declared dims of the *other* object in the ALIGN.

	If neither an 'AL_EXPL_' or 'AL_DEDUCED_' flag is set for the
source or target, it means that no list was supplied for it in ALIGN,
but either:
-- it was deduced to be (:,:,..) from that object's *known, declared*
rank, or
-- it's assumed to be scalar for the time being.
----------------------------------------------------------------*/
#define  AL_EXPL_SRC_LIST	0x4000000L
#define  AL_DEDUCED_SRC_LIST	0x8000000L

#define  AL_EXPL_TGT_LIST	0x10000000L
#define  AL_DEDUCED_TGT_LIST	0x20000000L
/*}}}*/

/*--------------------------------------------------------------*/
/* The following flag means 'descriptive' (i.e. '*') ALIGN	*/
/*--------------------------------------------------------------*/
#define  AL_ASSERT		0x40000000L

/*--------------------------------------------------------------*/
/* The following flag means that 'check_align_target_vs_dims ()'*/
/* has checked that 'tgt_rank' is correct and 'fixed' it if not	*/
/*--------------------------------------------------------------*/
#define  AL_CHECKED_TGT		0x80000000L


/*--------------------------------------------------------------*/
/*	The next 2 flags are used only while 'yyparse()' is	*/
/* reading the align source and target lists into buffers	*/
/* 'ep1_buf []' & 'ep2_buf []'.  They're cleared before		*/
/* 'install_align ()' sets-up the 'alignp' record properly.	*/
/*	During their lifetime they only co-exist with flags in	*/
/* the set:							*/
/*	(AL_SRC_RANK_MASK | AL_TGT_RANK_MASK |			*/
/*	 AL_EXPL_SRC_LIST | AL_EXPL_TGT_LIST | AL_ASSERT),	*/
/* so they can overlap any of the other flags above.		*/
/*--------------------------------------------------------------*/
#define  AL_TOO_MANY_SRC_DIMS	AL_NEED_SRC_LB	/* i.e. src list >7 dims */
#define  AL_TOO_MANY_TGT_DIMS	AL_NEED_TGT_LB	/* i.e. tgt list >7 dims */
/*}}}*/
/*{{{  'distp->attrs' flags  ('DI_')*/
/*----------------------------------------------*/
/* Flags for 'distp->attrs'			*/
/*						*/
/* N.B. This assumes that 'int32' is 'long'	*/
/*----------------------------------------------*/
/*{{{  simple flags*/
/*--------------------------------------------------------------*/
/* The following fields must accommodate values <= MAX_RANK,	*/
/* i.e. 7 with current Fortran rules, so they're 3 bits wide.	*/
/*								*/
/* N.B.:							*/
/*	proc_rank = #(BLOCK / CYCLIC entries in 'distp->fmt []').*/
/* This should equal the rank of the target procs arrangement.	*/
/* If it doesn't, then either the format list (i.e. 'fmt[]' and	*/
/* 'b[]') or the target processors arrangement 'proc_sp' is	*/
/* changed to correct the situation.				*/
/*--------------------------------------------------------------*/
#define  DI_SRC_RANK_MASK	0x7L		/* field for 'src_rank' */
#define  DI_PROC_RANK_MASK	0x38L		/*   "    "  'proc_rank */

/*----------------------------------------------------------------
'xxx_SHIFT' is the number of bits that a value must be shifted
(left) to reach the field in which it's stored in 'attrs'.
E.g., 'proc_rank' is stored and read as follows:
   to store:	attrs    += proc_rank << DI_PROC_RANK_SHIFT
   to read:	proc_rank = (attrs & DI_PROC_RANK_MASK) >> DI_PROC_RANK_SHIFT
----------------------------------------------------------------*/
#define  DI_PROC_RANK_SHIFT	3


/*----------------*/
/* Other flags... */
/*----------------*/
#define  DI_EXPL_FMT_LIST	0x40L
#define  DI_DEDUCED_FMT_LIST	0x80L
#define  DI_DEDUCED_PROC_RANK	0x100L

#define  DI_ASSERT_FMT		0x200L
#define  DI_ASSERT_PROCS	0x400L
#define  DI_INHERIT_FMT		0x800L
#define  DI_INHERIT_PROCS	0x1000L

#define  DI_TOO_MANY_SRC_DIMS	0x2000L

#define  DI_CHECKED_PROC_TYPE	0x4000L
#define  DI_CHECKED_PROC_DIMS	0x8000L

/*----------------------------------------------------------------
	Flag '(DI_NEED_BLOCKSIZE << dim)',  0 <= dim < MAX_RANK,
is set if:
	((fmt [dim] & DFMT_BLOCK)  &&  (b [dim] == NULL))
meaning that dimension 'dim' has BLOCK distribution with default
blocksize  (which must be evaluated and set when the dimensions
of both 'sp' and 'proc_sp' are known).
	'DI_NEED_BLOCKSIZE_MASK' is the mask for the set of flags
{(DI_NEED_BLOCKSIZE << dim),  0 <= dim < MAX_RANK},
----------------------------------------------------------------*/
#define  DI_NEED_BLOCKSIZE	 0x10000L
#define  DI_NEED_BLOCKSIZE_MASK	 0x7F0000L


/*}}}*/

/*{{{  composite flags*/

#define  DI_ASSERT		(DI_ASSERT_FMT | DI_ASSERT_PROCS)
#define  DI_INHERIT		(DI_INHERIT_FMT | DI_INHERIT_PROCS)
/*}}}*/
/*}}}*/
/*{{{  'distp->fmt []' flags ('DFMT_...')*/
/*------------------------------*/
/* Flags for 'distp->fmt []'	*/
/*------------------------------*/
#define  DFMT_COLLAPSED		0x1
#define  DFMT_BLOCK		0x2
#define  DFMT_CYCLIC		0x4
#define  DFMT_INHERIT		0x8
/*}}}*/
/*{{{  'mapp->attrs' flags   ('MA_...')*/
/*----------------------------------------------*/
/* Flags for 'mapp->attrs'			*/
/*						*/
/* N.B. This assumes that 'int32' is 'long'.	*/
/*----------------------------------------------*/
/*{{{  simple flags*/
/*--------------------------------------*/
/* Simle flags for 'mapp->attrs'	*/
/*--------------------------------------*/
/*-----------------------------------------*/
/* data types (**SAME AS FOR 'sp->type'**) */
/*-----------------------------------------*/
#define  MA_KIND_MASK		TY_KIND_MASK	/* 0x3FL */

#define  MA_INTEGER		TY_INTEGER	/* 0x40L */
#define  MA_REAL		TY_REAL		/* 0x80L */
#define  MA_DP			TY_DP		/* 0x100L */
#define  MA_COMPLEX		TY_COMPLEX	/* 0x200L */
#define  MA_LOGICAL		TY_LOGICAL	/* 0x400L */
#define  MA_CHARACTER		TY_CHARACTER	/* 0x800L */
#define  MA_DERIVED		TY_DERIVED	/* 0x1000L */
#define  MA_NONE		TY_NONE		/* 0x2000L */

/*--------------------------------------------------------------*/
/* The remaining flags are different from 'sp->type' flags...	*/
/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/
/* The following fields must accommodate values <= MAX_RANK,	*/
/* i.e. 7 with current Fortran rules, so they're 3 bits wide.	*/
/*--------------------------------------------------------------*/
#define  MA_ARRAY_RANK_MASK	0x1C000L	/* field for 'array_rank' */
#define  MA_EMBED_RANK_MASK	0xE0000L	/*   "    "  'embed_rank'*/

/*--------------------------------------------------------------*/
/* 'xxx_SHIFT' is the number of bits that a value must be	*/
/* shifted (left) to reach its field in 'attrs'.		*/
/*--------------------------------------------------------------*/
#define  MA_ARRAY_RANK_SHIFT	14		/* shift for 'array_rank' */
#define  MA_EMBED_RANK_SHIFT	17		/*   "    "  'embed_rank'*/

/*----------------------------------------------*/
/* one of the next 3 flags is always set...	*/
/*** N.B.  REORDER THEIR VALUES AFTER TELLING DBC !!! ********/
/*----------------------------------------------*/
#define  MA_DISTRIB		0x800000L	/* distributed */
#define  MA_GLOBALLY_REPLIC	0x1000000L	/* globally replicated
						      (proc_sp == NULL) */
#define  MA_REPLIC_ON_PROCS	0x20000000L	/* replicated over 'proc_sp'
						      (proc_sp != NULL)*/
/*--------------------------------------------------------*/
/* if any of the next 4 flags are set, so is 'MA_DISTRIB' */
/*--------------------------------------------------------*/
#define  MA_INHERITED		0x2000000L	/* "INHERIT" -- nothing
						    known about mapping! */
#define MA_ALIGNED_WITH_INHERIT	0x4000000L	/* "ALIGN -> INHERIT" */

#define  MA_INHERIT_FMT		0x8000000L	/* "DISTRIBUTE *" */
#define  MA_INHERIT_PROCS	0x10000000L	/* "DISTRIBUTE ... ONTO *" */

/*--------------------------------------------------------*/
/* 'mapp' for an assumed (or deferred!) shape array	*/
/*--------------------------------------------------------*/
#define  MA_ASSUMED_SHAPE	0x40000000L

/*}}}*/

/*{{{  composite flags*/
/*-----------------*/
/* Composite flags */
/*-----------------*/
/*-- datatype masks (** SAME AS FOR 'sp->type'**) --*/

#define  MA_BASIC_TYPE		TY_BASIC_TYPE	/* type *without* KIND val */
#define  MA_TYPE		TY_TYPE		/* type *with* KIND val */

#define  MA_INHERIT_MASK	(MA_INHERITED | MA_ALIGNED_WITH_INHERIT |\
				 MA_INHERIT_FMT | MA_INHERIT_PROCS)
/*}}}*/
/*}}}*/

/*{{{  temp flags, used in both 'arp->attrs' & 'distp->attrs'*/
/*--------------------------------------------------------------*/
/* The next flags are temporary -- they may only be set during	*/
/* the processing of a 'shape_spec' or 'distr_spec', after	*/
/* which they're deleted.  The first 2 flags are used in both	*/
/* 'arp->attrs' and 'distp->attrs'.				*/
/*--------------------------------------------------------------*/
#define  NOT_SPEC_EXPR	0x4000	/* some bounds (or procdims) aren't spec exprs. */
#define  NEG_PROCDIMS	0x8000	/* some procdims are -ve */

/*}}}*/

/*{{{  'exprtype' ******* COMMENTED OUT -- DELETE IT?? ********/
/*--------------*/
/*  'exprtype'	*/
/*--------------*/
/****************************************
A BIG PROBLEM WITH THIS VERSION  ---
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
It mixes type and attr flags, which cannot be mixed as they are overlapping.

#define  INIT_EXPR	(TY_NUMERIC | AT_PARAM)
#define  IMPDO_EXPR	(TY_INTEGER | (AT_PARAM | AT_NORMAL_VAR))
#define  SPEC_EXPR	(TY_INTEGER | AT_DEF_ON_ENTRY ***| AT_EVAL***)

*-- 'attrs' field in 'exprtype' --*

#define  AT_EVAL	AT_ARRAY * integer literal expression;	*
				 * can be evaluated by YACC	*

******************************************/
/*}}}*/

#define  MAX_RANK	7

#define  NULL_Sym	((Symbol *) NULL)
#define  NULL_Array	((Array *) NULL)
#define  NULL_Align	((Align *) NULL)
#define  NULL_Distr	((Distr *) NULL)
#define  NULL_Map	((Map *) NULL)
#define  NULL_Map_p	((Map **) NULL)

#define  HASHSIZE	101
/*}}}*/
/*{{{  typedefs  (Typeinfo, Symbol, Array, Align, Distr, Map)*/
/*{{{  struct Typeinfo*/
/*--------------------------------------------------------------*/
/* Union 'Typeinfo' is only used in the 'Symbol' and 'Unit'	*/
/* structs.  It stores 'extra' information required for certain	*/
/* types, viz: CHARACTER*(len) and TYPE (typename)		*/
/*--------------------------------------------------------------*/
typedef union {
	struct Expr   *char_len;	/* for CHARACTER*len */
	struct Symbol *type_sp;		/* for deriv type: ptr to typedef */
} Typeinfo;
/*}}}*/

/*{{{  struct Symbol*/
/*--------------------------------------------------------------*/
/* Symbol (only for storing user IDs).				*/
/*								*/
/* N.B.  Members 'name', 'attrs', 'type' must be in positions	*/
/* 1 - 3 respectively, as that order is assumed by the		*/
/* initialiser of array 'Symbol intrin_procs[]' in "intrin.c".	*/
/*--------------------------------------------------------------*/
typedef struct Symbol {  
	char     *name;
	int32     attrs;		/* see above for flags */
	int32     type;			/* type & KIND; see above for flags*/
	Typeinfo  typeinfo;		/* for CHAR*(len) or TYPE (name) */
	short     scope;		/* 'scope number' */
	struct Array   *arp;		/* ptr to DIMENSION information */
	struct Distr   *distp;
	struct Align   *alignp;
	struct Map     *mapp;
	union {
		int     i_init;		/* simple integer */
		char   *c_init;		/* character string */
		struct Expr    *init;	/* general expression */
		struct Symbol  *sp;	/* e.g. for proc interface */
		int   (*fn)();		/* symbol handling fn (eg, for intrin procs) */
	} info;
#ifdef V_IDA
	Idref    *first_refp;		/* first ref, for IDA */
	Idref    *last_refp;		/* last ref, for IDA */
#endif
	struct Expr    *actual_arg;	/* only for dummy args in proc interface */
	
	struct Symbol   *next;
	struct Symbol   *nextid;
} Symbol;
/*}}}*/

/*{{{  struct Array*/
/*--------------------------------------------------------------*/
/* Array							*/
/*								*/
/* N.B.  'attrs' must be the first member, as that's assumed by	*/
/* the initialisers for 'dummy_arp', etc, in "intrin.c".	*/
/*--------------------------------------------------------------*/
typedef struct Array {
	int           attrs;
	struct Expr  *l [MAX_RANK];
	struct Expr  *n [MAX_RANK];	/* NULL for assumed shape */
	short         nlinks;
} Array;
/*}}}*/

/*{{{  struct Align*/
/*--------*/
/* Align  */
/*--------*/
typedef struct Align {
	Symbol        *tgt_sp;	/* NULL if ultimate tgt is 'INHERITed' */
	int32          attrs;
	short          nlinks;

	struct Expr  **al;	/* al [src_rank] : alignment lower bounds */
	struct Expr  **as;	/* as [src_rank] : alignment strides */
	short         *d;	/* d  [src_rank] : align target dims */
	/*------------------------------------------------------*/
	/* The following are omitted for INHERIT, as the	*/
	/* 'embed_rank' & 'replic_rank' are unknown statically..*/
	/*------------------------------------------------------*/
	struct Expr  **ei;	/* ei [embed_rank] : subscripts in
				                    embedded target dims */
	short         *ed;	/* ed [embed_rank] : embedded target dims */
	short         *rd;	/* rd [replic_rank] : replicated tgt dims */
} Align;
/*}}}*/

/*{{{  struct Distr*/
/*--------*/
/* Distr  */
/*--------*/
typedef struct Distr {
	Symbol        *proc_sp;	/* defined unless (attrs & DI_INHERIT_PROCS) */
	int32          attrs;
	short          nlinks;

	short         *fmt;	/* fmt [src_rank] : distribution formats */
	struct Expr  **b;	/* b [src_rank] :   block sizes;  NULL if
				        & only if (fmt == DFMT_COLLAPSED) */
} Distr;
/*}}}*/

/*{{{  struct Map*/
/*--------------------------------------------------------------*/
/* Map:  a structure to record HPF mapping in a 'rationalised'	*/
/* form in both the symbol table ('sp->mapp') and expression	*/
/* nodes (ep->mapp).						*/
/*--------------------------------------------------------------*/
typedef struct Map {
	int32  attrs;		/* gives type, array_/embed_/replic_rank */
	short  nlinks;
	struct Expr  *mapvec_ep;	/* expr for object's 'mapvec' */
	/*------------------------------------------------------*/
	/* Vectors of length 'array_rank'. 			*/
	/*							*/
	/* N.B.   Logically there are 7 of them.  However, in	*/
	/* fact they're allocated as 2 vectors:			*/
	/* -- *al: a vector of length [5 *array_rank] containing*/
	/*         (al[], an[], as[], tn[], b[]);		*/
	/* -- *fmt: a vector of size [2 *array_rank] containing	*/
	/*          (fmt[], d[]).				*/
	/*------------------------------------------------------*/
	struct Expr  **al,  **an,  **as,  **tn,  **b;
	short         *fmt,  *d;

	/*-----------------------------------*/
	/* vectors of length 'embed_rank'... */
	/*-----------------------------------*/
	struct Expr  **ep;
	short         *ed;

	struct Symbol  *proc_sp;
} Map;
/*}}}*/
/*}}}*/
/*{{{  function decls*/
/*------------------*/
/*  function decls  */
/*------------------*/
unsigned hash ();
Symbol *lookup_or_install_id ();
bool	canbe_aligned (),  canbe_dimensioned (),  canbe_distributed (),
	check_idclass_in_expr (),  check_num_in_expr (),
	mustbe_scalar_var (),  mustbe_spec_expr ();
char   *ecalloc (),  *emalloc ();
Array  *create_array ();
Align  *create_align ();
Distr  *create_distr ();
/*}}}*/
/*{{{  macros for flag manipulation*/

#define  undeclared_id(sp)		(!((sp)->attrs || (sp)->type ||	\
					   (sp)->arp || (sp)->alignp ||	\
					   (sp)->distp))

#define  only_flags_set(word,flags)	(((word) & (flags)) == (word))
#define  all_flags_set(word,flags)	(((word) & (flags)) == (flags))
#define  clear_flags(word,flags)	((word) &= ~(flags))
#define  not_set(word,flags)		(! ((word) & (flags)))
#define  any_entity_except(word,flags)	((word) & (AT_ENTITY & ~(flags)))
#define  change_flags(word,from_flags,to_flags)				\
			((word) &= ~(from_flags), (word) |= (to_flags))

/*}}}*/
/*{{{  externs*/
/*------------*/
/*  globals  */
/*------------*/
extern Symbol	*last_sp;
/*}}}*/


#endif
